from typing import Iterable, Union, Type
from . import _Table
from . import _Where
from . import _internal
from . import _exceptions


class UpdateQuery:
    """
    UpdateQuert object to perform update on table.
    """
    def __init__(self, selection: 'SelectQuery', values: tuple | dict):
        self._target = selection.source
        self._fields = selection.fields
        self._values = values
        self._where = selection.where_condition
        self._having = selection.having_condition
        self._group = selection.group_fields

        if (not self._target.is_real) and (self._where is not None):
            raise _exceptions.TableIsNotReal(self._target.name)

        if not isinstance(values, dict) and ((lv := len(values)) != (lf := len(self._fields))):
            raise ValueError(f'All fields from selection have to be initialized in UPDATE query, but only {lv} of {lf} are.')

    def __call__(self, commit=True) -> None:
        """
        Make SQL UPDATE Query, which is presented by current object.

        :param commit: True means to commit changes to database after query success.
        """
        query = str(self)
        self._target.db.query(query, commit=commit)

    def __str__(self) -> str:
        if isinstance(self._values, tuple):
            fields = (f.name for f in self._fields)
            values = self._values
        elif isinstance(self._values, dict):
            fields = (self._target.field_by_name(f).name for f in self._values.keys())
            values = self._values.values()
        else:
            raise TypeError(f'expected <dict> or <tuple> for value assignment, got {type(self._values)}')

        update = f'''UPDATE
        {self._target.name} SET {",".join(fields)} = ({",".join(_internal.proper_values(values))})'''
        where = f' WHERE {self._where}' if self._where is not None else ""
        group = f' GROUP BY {",".join(f.full_name for f in self._group)}' if self._group else ""
        having = f' HAVING {self._having}' if self._having else ""
        return f'{update}{where}{group}{having};'

    def __repr__(self) -> str:
        return str(self)


class SelectQuery:
    """
    SelectQuery object to perform SELECT queries and its compositions.
    """
    def __init__(self, source: '_Table.Table', fields: Iterable['_Table.TableField']):
        self._source = source
        self._fields = tuple(fields)
        self._distinct = False
        self._where = None
        self._having = None
        self._body = None
        self._union_comparator = _Where.WhereAND
        self._group: tuple['_Table.TableField'] = tuple()
        self._order: tuple['_Table.TableField'] = tuple()


    @property
    def union_comparator(self) -> Type['_Where.WhereComposition']:
        """Comparator type which will be used on further conditions composition for this query."""
        return self._union_comparator

    @property
    def is_distinct(self) -> bool:
        """Is this query DISTINCT or not."""
        return self._distinct

    @property
    def where_condition(self) -> Type['_Where.Where']:
        return self._where

    @property
    def having_condition(self) -> Type['_Where.Where']:
        return self._having

    @property
    def group_fields(self) -> tuple['_Table.TableField']:
        return self._group

    @property
    def order_fields(self) -> tuple['_Table.TableField']:
        return self._order

    @property
    def source(self) -> '_Table.Table':
        """Source Table object to SELECT from."""
        return self._source

    @property
    def fields(self) -> tuple['_Table.TableField']:
        """Fields used in current selection."""
        return self._fields

    @property
    def body(self):
        """Last query operation result. Used for caching."""
        return self._body

    @property
    def distinct(self) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but selection will be DISTINCT."""
        new = self.copy()
        new._distinct = True
        return new

    @property
    def OR(self) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but newly added selection conditions will be added with OR."""
        new = self.copy()
        new._union_comparator = _Where.WhereOR
        return new

    @property
    def AND(self) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but newly added selection conditions will be added with AND."""
        new = self.copy()
        new._union_comparator = _Where.WhereAND
        return new

    def WHERE(self, values: list | tuple, comparison: type, union: type) -> 'SelectQuery':
        """
        Make new SelectQuery, which is copy of current, but add new selection condition.

        :param values: condition description.
        :param comparison: Where class reference which represents how values would be checked.
        :param union: WhereComposition class reference which represents how all checks will be composited.
        :return: new SelectQuery object.
        """
        if isinstance(values, list):
            where_type = _Where.WhereAND
        elif isinstance(values, tuple):
            where_type = _Where.WhereOR
        else:
            raise TypeError(f'expected <tuple> for AND or <list> for OR, got {type(values)}')

        new_wheres = []
        new_havings = []
        for value, field in zip(_internal.proper_values(values), self.fields):
            if (value == tuple()) or (value == list()):
                continue
            if isinstance(value, tuple):
                new_condition = _Where.WhereOR(*(comparison(field, v) for v in _internal.proper_values(value)))
            elif isinstance(value, list):
                new_condition = _Where.WhereAND(*(comparison(field, v) for v in _internal.proper_values(value)))
            else:
                new_condition = comparison(field, value)

            if isinstance(field, _Table.CalculatedField):
                new_havings.append(new_condition)
            else:
                new_wheres.append(new_condition)

        where = union(self._where, where_type(*new_wheres)) if self._where is not None else where_type(*new_wheres)
        having = union(self._having, where_type(*new_havings)) if self._having is not None else where_type(*new_havings)

        new = self.copy()
        new._where = where if new_wheres else None
        new._having = having if new_havings else None
        return new

    def WHERE_EQ(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on equalities."""
        return self.WHERE(values, _Where.WhereEq, self._union_comparator)

    def WHERE_GT(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on greaters."""
        return self.WHERE(values, _Where.WhereGt, self._union_comparator)

    def WHERE_LT(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on lessers."""
        return self.WHERE(values, _Where.WhereLt, self._union_comparator)

    def GROUPBY(self, fields: tuple) -> 'SelectQuery':
        """Make new SelectQuery, which copy of current, but with grouping fields."""
        fields = tuple(self._source.field_by_name(v) for v in fields)
        new = self.copy()
        new._group = fields
        return new

    def ORDERBY(self, fields: tuple) -> 'SelectQuery':
        fields = tuple(self._source.field_by_name(v) for v in fields)
        new = self.copy()
        new._order = fields
        return new

    def UPDATE(self, values: tuple | list) -> 'UpdateQuery':
        """Make new UpdateQuery, that affects rows and fields selected with current SelectQuery."""
        return UpdateQuery(self, values)

    def __eq__(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on equalities."""
        return self.WHERE_EQ(values)

    def __gt__(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on greaters."""
        return self.WHERE_GT(values)

    def __lt__(self, values: list | tuple) -> 'SelectQuery':
        """Make new SelectQuery, which is copy of current, but with additional selection condition on lessers."""
        return self.WHERE_LT(values)

    def __mod__(self, fields: tuple) -> 'SelectQuery':
        """Make new SelectQuery, which copy of current, but with grouping fields."""
        return self.GROUPBY(fields)

    def __div__(self, fields: tuple) -> 'SelectQuery':
        return self.ORDERBY(fields)

    def __lshift__(self, values: tuple | dict) -> 'UpdateQuery':
        """Make new UpdateQuery, that affects rows and fields selected with current SelectQuery."""
        return self.UPDATE(values)

    def __call__(self) -> list[tuple]:
        """
        Get result of SQL query, which is presented by current object.

        :return: select SQL query result.
        """
        query = str(self)
        self._body = self._source.db.query(query, commit=False).fetchall()

        return self._body

    def __str__(self):
        select = f'''SELECT {"DISTINCT" if self._distinct else ""}
        {",".join(f.full_name for f in self.fields)} 
        FROM {self.source.query}'''
        where = f' WHERE {self._where}' if self._where else ""
        group = f' GROUP BY {",".join(f.full_name for f in self._group)}' if self._group else ""
        order = f' ORDER BY {",".join(f.full_name for f in self._order)}' if self._order else ""
        having = f' HAVING {self._having}' if self._having else ""

        query = f'{select}{where}{group}{having}{order};'
        return query

    def __repr__(self) -> str:
        return str(self)

    def copy(self) -> 'SelectQuery':
        new = SelectQuery(self._source, self._fields)
        new._distinct = self._distinct
        new._where = self._where
        new._having = self._having
        new._group = self._group
        new._order = self._order
        new._union_comparator = self._union_comparator

        return new



from ._internal import *


def transfer_object(cls: ClassVar):
    fields = public_fields(cls)
    for name, value in fields.items():
        if (cname := contr_name(name)) not in cls.__dict__:
            if isinstance(value, tuple) and len(value) >= 2:
                setattr(cls, cname, value[0])
                setattr(cls, name, value[1])
            else:
                setattr(cls, cname, f'{name}')
                setattr(cls, name, value)
        else:
            setattr(cls, name, value)

        if (ctype := contr_type(name)) not in cls.__dict__:
            setattr(cls, ctype, type(cls.__dict__[name]))

        if isinstance(value, tuple) and len(value) >= 3:
            setattr(cls, allowed_values(name), value[2])
        elif (alname := allowed_values(name)) not in cls.__dict__:
            setattr(cls, alname, tuple())

    cls.get_contract = get_contract
    cls.as_dict = public_fields
    cls.set_value = set_value
    cls.write_tuple = write_tuple
    cls.__setattr__ = set_value
    cls.__init__ = contract_init

    return cls






class FieldConstraint:
    """Field constraint base class"""
    name = ""

    def __init__(self, *args):
        self.details = [str(a) for a in args]

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return f'{self.name}({",".join(self.details)})'

    def __hash__(self):
        return hash((self.name, self.details))

    def __eq__(self, other: 'FieldConstraint'):
        return hash(self) == hash(other)


class NotNull(FieldConstraint):
    """Not NULL constraint"""
    name = "NOTNULL"


class Primary(FieldConstraint):
    """Primary key constraint"""
    name = "PRIMARY"


class Foreign(FieldConstraint):
    """Foreign key constraint"""
    name = "FOREIGN"


class Check(FieldConstraint):
    name = "CHECK"


class Unique(FieldConstraint):
    name = "UNIQUE"


class Default(FieldConstraint):
    name = "DEFAULT"


from abc import abstractmethod
from typing import TypeVar, Protocol


class Equatable(Protocol):
    @abstractmethod
    def __eq__(self: 'EqType', other: 'EqType') -> bool:
        pass


class Hashable(Protocol):
    @abstractmethod
    def __hash__(self: 'HashType'):
        pass


class Comparable(Protocol):
    @abstractmethod
    def __lt__(self: 'CompType', other: 'CompType') -> bool:
        pass

    @abstractmethod
    def __gt__(self: 'CompType', other: 'CompType') -> bool:
        pass

    @abstractmethod
    def __eq__(self: 'CompType', other: 'CompType') -> bool:
        pass


CompType = TypeVar('CompType', bound=Comparable)
EqType = TypeVar('EqType', bound=Equatable)
HashType = TypeVar('HashType', bound=Hashable)



import sqlite3
from typing import Iterable, Union

from . import _exceptions
from . import _Table
from . import _internal
from . import _FieldConstraints as _Constr


class DBase:
    """
    SQLite3 database class.
    """
    def __init__(self, dbfile: str):
        self._db_file = dbfile

        self._db_connection = sqlite3.connect(self._db_file)
        self._db_cursor = self._db_connection.cursor()

        self._db_tables = self.get_tables()
        self._active_tables: dict[str, '_Table.Table'] = dict()

    def table(self, table_name: str) -> '_Table.Table':
        """
        Get Table from base.

        Get interactable Table object from this DBase that represents table from base.

        If table with given name have already been initialized with Table object, reference to if will be returned, not new Table object.

        :param table_name: Name of existing table in sqlite3 database.
        :return: new Table object or reference to existing Table object.
        """
        if self.has_table(table_name):
            if table_name in self._active_tables:
                table = self._active_tables[table_name]
            else:
                table = _Table.Table(table_name, self)
                self._active_tables[table_name] = table
            return table
        else:
            raise _exceptions.TableNotFound(table_name, self.name)

    def new_table(self, table_name: str, fields: dict) -> '_Table.Table':
        """
        Create new table in database and get reference Table object for it.

        If table already exists in database, exception will be thrown.

        :param table_name: new table name.
        :param fields: schema for new table.
        :return: new Table object that represents newly created table.
        """
        if self.has_table(table_name):
            raise _exceptions.TableAlreadyExists(table_name, self.name)
        else:
            fields = f'{", ".join(f"{k} {v}" for k, v in fields.items())}'
            query = f'CREATE TABLE {table_name}({fields});'
            self.query(query)
            self._db_connection.commit()

            self._db_tables = self.get_tables()

            return self.table(table_name)

    def query(self, query: str, commit=False):
        """
        Make string SQL query to database.

        :param query: SQL query string.
        :param commit: Either commit changes to database with this query or not.
        :return: query result whatever is is.
        """
        res = self._db_cursor.execute(query)
        if commit:
            self._db_connection.commit()
        return res

    def select(self, source: str, fields: Iterable[str]) -> list[tuple]:
        """
        Select fields from table with no confines.

        :param source: source string name as it is in database.
        :param fields: iterable collection of field string names.
        :return: list of tuples with selected values
        """
        return self.query(f'SELECT {",".join(fields)} FROM {source};').fetchall()

    def insert(self, target_table: str, target_fields: Iterable[str], values: Iterable) -> None:
        """
        Insert values into target table.

        :param target_table: target table string name as it is in database.
        :param target_fields: iterable collection of target fields
        :param values: iterable collection of target values
        :return:
        """
        f_names = []
        f_values = []
        for f, v in zip(target_fields, _internal.proper_values(values)):
            if v == "NULL":
                continue
            f_names.append(f)
            f_values.append(v)

        self.query(f'INSERT INTO {target_table}({",".join(f_names)}) VALUES({",".join(f_values)});', commit=True)
        self._db_connection.commit()

    def drop(self, target_table: str, suppress_nonexisting=False) -> None:
        """
        Drop table from current DBase.

        :param target_table: target table name as it is in sqlite base
        :param suppress_nonexisting: True -- add 'IF EXISTS', so no error on table is not in db
        """
        self.query(f'DROP TABLE {target_table} {"IF EXISTS" if suppress_nonexisting else ""};')

    def has_tables(self, tables: Iterable[Union[str, '_Table.Table']]) -> bool:
        """
        Check if all given tables are present in database.

        :param tables: iterable collection of string table names or of Table objects to check
        :return: True if all tables are present
        """
        for table in tables:
            if not self.has_table(table):
                return False
        return True

    def has_table(self, table_name: Union[str,'_Table.Table']) -> bool:
        """
        Check if given table is present in database.

        :param table_name: string name of table of Table object to check
        :return: True if table is present
        """
        table_name = table_name if isinstance(table_name, str) else table_name.name
        return table_name in self._db_tables

    def get_tables(self) -> set[str]:
        """
        Get all tables from database with SQL query.

        :return: set of table names
        """
        tables = self.query('SELECT name from sqlite_master where type="table"').fetchall()
        return set(t[0] for t in tables)

    def get_fields(self, table: str) -> dict[str, '_Table.TableField']:
        """
        Get fields for of given table.

        If table is not present, exception will be thrown.

        Also, when call this method, table will be registered in DBase and Table object for given table will be created.

        :param table: string name of table as it is in database.
        :return: dict of {field_name: TableField object}
        """
        if not self.has_table(table):
            raise _exceptions.TableNotFound(table, self.name)

        table = self.table(table)
        return self.table_fields(table)

    def get_foreign_keys(self, table: str) -> dict[str, '_Table.TableFK']:
        """
        Get foreign keys of given table.

        If table is not present, exception will be thrown.

        Also, when call this method, table will be registered in DBase and Table object for given table will be created.

        :param table: table string name as it is in base.
        :return: dict of {field_name: TableFK}
        """
        if not self.has_table(table):
            raise _exceptions.TableNotFound(table, self.name)

        table = self.table(table)
        return self.table_foreign_keys(table)

    def table_satisfied(self, table: '_Table.Table') -> bool:
        """
        Check if all tables binded to some Table object are present in database.

        Useful for non-real Tables.

        :param table: Table object reference.
        :return: True if all binded tables are present in base.
        """
        return self.has_tables(table.binded)

    def tables_satified(self, tables: Iterable['_Table.Table']) -> bool:
        """
        Check if all tables binded to all given Table objects are present in database.

        Useful for non-real Tables.

        :param tables: Iterable of Table object references.
        :return: True if all binded tables for all given Table objects are present in base.
        """
        for t in tables:
            return self.table_satisfied(t)
        return True

    def table_fields(self, table: '_Table.Table') -> dict[str, '_Table.TableField']:
        """
        Get fields for of given table.

        :param table: Table reference object.
        :return: dict of {field_name: TableField object}
        """
        result = self.query(f'PRAGMA table_info({table.query})').fetchall()
        fks = self.query(f'PRAGMA foreign_key_list({table.query})').fetchall()
        fields = dict()
        for i, name, typ, nullable, default, pk in result:
            constraints = []
            if pk != 0:
                constraints.append(_Constr.Primary(pk))
            if not nullable:
                constraints.append(_Constr.NotNull())
            if default:
                constraints.append(_Constr.Default(default))
            fields.update({f"{table.name}.{name}": _Table.TableField(i, name, typ, table, constraints=constraints)})

        for fk in fks:
            master_name = fk[2]
            master_field = fk[4]
            slave_field = fk[3]
            fields[f'{table.name}.{slave_field}'].constraints.append(_Constr.Foreign(f'{master_name}.{master_field}'))

        return fields

    def table_foreign_keys(self, table: '_Table.Table') -> dict[str, '_Table.TableFK']:
        """
        Get foreign keys for of given table.

        :param table: Table reference object.
        :return: dict of {field_name: TableFK object}
        """
        result = self.query(f'PRAGMA foreign_key_list({table.query})').fetchall()
        keys = dict()

        for f in result:
            master = self.table(f[2])
            fk = _Table.TableFK(master.field_by_name(f[4]), table.field_by_name(f[3]))
            keys[f'{table.name}.{fk.slave_field.name}'] = fk

        return keys

    @property
    def name(self) -> str:
        """Name of current DBase's file."""
        return self._db_file

    @property
    def tables(self) -> set[str]:
        """All tables present in current DBase's file."""
        return self._db_tables

    @property
    def active_tables(self) -> dict[str, '_Table.Table']:
        """All Table objects initialized within DBase."""
        return self._active_tables.copy()

    def __repr__(self) -> str:
        return f'DBase<{self.name}>'

    def __del__(self):
        self._db_cursor.close()
        self._db_connection.close()



from typing import Hashable, Iterable
from math import inf as INF

from ..primitives import Queue


class PyGraph:
    def __init__(self):
        self._graph_vertex: set[Hashable] = set()
        self._graph_adj: dict[Hashable, set] = dict()
        self._adj_weights: dict[tuple[Hashable, Hashable], float] = dict()
        self._vertex_attr: dict[Hashable, set] = dict()

    def add_vertex(self, value: Hashable):
        if value not in self:
            self._graph_adj[value] = set()
            self._vertex_attr[value] = set()
            self._graph_vertex.add(value)

    def add_edge(self, value1: Hashable, value2: Hashable, weight: float = 1):
        self.add_vertex(value1)
        self.add_vertex(value2)

        self._graph_adj[value1].add(value2)
        self._graph_adj[value2].add(value1)
        self._adj_weights[(value1, value2)] = weight
        self._adj_weights[(value2, value1)] = weight

    def edge_weight(self, value1: Hashable, value2: Hashable) -> float:
        if (value1, value2) in self._adj_weights:
            return self._adj_weights[(value1, value2)]
        else:
            return INF

    def are_adjacent(self, value1: Hashable, value2: Hashable) -> bool:
        if value1 in self and value2 in self:
            return (value2 in self._graph_adj[value1]) and (value1 in self._graph_adj[value2])
        else:
            raise ValueError(f"`{value1}` or `{value2}` not in graph")

    def adjacents(self, value: Hashable) -> set:
        if value in self:
            return self._graph_adj[value]
        else:
            return None

    def travel(self, start_value: Hashable) -> Iterable[Hashable]:
        if start_value not in self:
            raise ValueError(f"`{start_value}` not in graph")

        visited = set()
        queue = Queue()
        queue.add(start_value)
        while not queue.is_empty():
            value = queue.next()
            visited.add(value)
            yield value
            for next_value in self._graph_adj[value]:
                if next_value not in visited:
                    queue.add(next_value)

    def travel_levels(self, start_value: Hashable) -> Iterable[tuple[Hashable, set[Hashable]]]:
        if start_value not in self:
            raise ValueError(f"`{start_value}` not in graph")

        visited = set()
        queue = [start_value]
        while queue:
            value = queue.pop(0)
            visited.add(value)
            yield value, self._graph_adj[value]
            for next_value in self._graph_adj[value]:
                if next_value not in visited:
                    queue.append(next_value)

    def minimal_path_deikstra(self, value_from: Hashable, value_to: Hashable) -> dict[str, list | float]:
        if not (value_from in self and value_to in self):
            raise ValueError(f"`{value_from}` or `{value_to}` not in graph")

        paths = dict()
        for v in self._graph_vertex:
            weight = self.edge_weight(value_from, v)
            path = [value_from, value_to] if weight < INF else []
            paths.update({v: {'path': path, 'weight': weight}})
        paths[value_from] = {'path': [value_from], 'weight': 0}
        unvisited = self._graph_vertex.copy()

        while unvisited:
            min_path = INF
            min_vertex = None
            for u in unvisited:
                if paths[u]['weight'] <= min_path:
                    min_path = paths[u]['weight']
                    min_vertex = u

            if min_path == INF:
                break

            unvisited.remove(min_vertex)
            for v, data in paths.items():
                if data['weight'] > (p := paths[min_vertex]['weight'] + self.edge_weight(min_vertex, v)):
                    paths[v]['weight'] = p
                    paths[v]['path'] = paths[min_vertex]['path'] + [min_vertex, v]
        return paths[value_to]

    def set_attr(self, vertex: Hashable, attr: Hashable):
        if vertex in self:
            self._vertex_attr[vertex].add(attr)
        else:
            raise ValueError(f"`{vertex}` not in graph")

    def del_attr(self, vertex: Hashable, attr: Hashable):
        if vertex in self:
            self._vertex_attr[vertex].remove(attr)
        else:
            raise ValueError(f"`{vertex}` not in graph")

    def get_attr(self, vertex: Hashable) -> set:
        if vertex in self:
            return self._vertex_attr[vertex]
        else:
            raise ValueError(f"`{vertex}` not in graph")

    def has_attr(self, vertex: Hashable, attr: Hashable) -> bool:
        if vertex in self:
            return attr in self._vertex_attr[vertex]
        else:
            raise ValueError(f"`{vertex}` not in graph")

    @property
    def vertexes(self) -> set:
        return self._graph_vertex.copy()

    def __contains__(self, value: Hashable) -> bool:
        return value in self._graph_vertex



from typing import ClassVar, List, Dict


def contract_init(self):
    cls = type(self)
    for field in cls.__dict__:
        if is_contract_field(cls, field):
            self.__dict__[field] = cls.__dict__[field]


def get_contract(self) -> List[dict]:
    contract = []
    cls = type(self)
    fields = public_fields(cls)
    for field, _ in fields.items():
        if is_contract_field(cls, field):
            contract.append({
                "field": field,
                "name": cls.__dict__[contr_name(field)],
                "type": cls.__dict__[contr_type(field)],
                "default": cls.__dict__[field],
                "value": self.__dict__[field],
                "allowed": cls.__dict__[allowed_values(field)]
            })
    return contract


def set_value(self, field, value):
    cls = type(self)
    if is_contract_field(cls, field):
        if isinstance(value, valid_type := cls.__dict__[contr_type(field)]):
            if is_allowed(cls, field, value):
                self.__dict__[field] = value
            else:
                raise ValueError(f"'{value}' in not allowed for '{field}' in {cls} contract. Allowed: {cls.__dict__[allowed_values(field)]}")
        else:
            raise TypeError(f"{type(value)} is not a vaild type for '{field}' in {cls} contract. Use {valid_type}")
    else:
        raise AttributeError(f"'{field}' is not a part of {cls} contract")


def is_contract_field(cls: ClassVar, name: str) -> bool:
    return name in cls.__dict__ and contr_name(name) in cls.__dict__ and contr_type(name) in cls.__dict__


def is_allowed(cls: ClassVar, field: str, value) -> bool:
    alfield = allowed_values(field)
    if cls.__dict__[alfield]:
        return value in cls.__dict__[alfield]
    else:
        return True


def write_tuple(self, values: tuple) -> None:
    fields = tuple(public_fields(self).keys())
    if len(fields) != len(values):
        raise ValueError(f"Value amount mismatch. Contract has {len(fields)}, tuple got {len(values)}")

    for i,v in enumerate(values):
        set_value(self, fields[i], v)


def public_fields(obj) -> dict:
    return dict(filter(lambda i: not i[0].startswith("_"), obj.__dict__.items()))


def contr_name(field: str) -> str:
    return f"_{field}_contract_name"


def contr_type(field: str) -> str:
    return f"_{field}_contract_type"


def allowed_values(field: str) -> str:
    return f"_{field}_contract_allowed"


from abc import abstractmethod
from typing import Protocol, TypeVar, Iterable


def proper_values(values: Iterable) -> list:
    """
    Tranfrosm given values to proper ones for SQL query.

    :param values: Iterable of Python values.
    :return: list of values configured for SQL query.
    """
    proper_values = []
    for v in values:
        if isinstance(v, str):
            proper_values.append(f'"{v}"')
        elif isinstance(v, list) or isinstance(v, tuple):
            proper_values.append(v)
        elif v is None:
            proper_values.append('NULL')
        else:
            proper_values.append(f'{v}')
    return proper_values


from typing import Iterable

from .pytrees._BinaryTree import BinaryTree as PyBinaryTree
from .pytrees._PrefixTree import PrefixTree as PyPrefixTree
from ._internal import CompType, HashType


class BinaryTree:
    def add(self, item: CompType):
        raise NotImplementedError()

    def is_empty(self) -> bool:
        raise NotImplementedError()

    def contains(self, item: CompType):
        raise NotImplementedError()

    @property
    def depth(self) -> int:
        raise NotImplementedError()


class PrefixTree:
    def add(self, sequence: Iterable[HashType]):
        raise NotImplementedError()

    def contains(self, sequence: Iterable[HashType]) -> bool:
        raise NotImplementedError()

    def count(self, sequence: Iterable[HashType]) -> int:
        raise NotImplementedError()


from .cpp_lib.dist.trees import BinaryTree, PrefixTree


from typing import Union

from . import _Table


class Where:
    """Where base class for all selection conditions."""
    _operator = "????"

    def __init__(self, field1: '_Table.TableField', field2: '_Table.TableField'):
        self._left = field1
        self._right = field2

        if isinstance(self._left, Union[_Table.TableField, _Table.CalculatedField]):
            self._left = self._left.full_name

        if isinstance(self._right, Union[_Table.TableField, _Table.CalculatedField]):
            self._right = self._right.full_name

    def __str__(self) -> str:
        return f'({self._left} {self.__class__._operator} {self._right})'


class WhereEq(Where):
    """Selection condition on equality ="""
    _operator = "="


class WhereGt(Where):
    """Selection condition on greater >"""
    _operator = ">"


class WhereLt(Where):
    """Selection condition on less <"""
    _operator = "<"


class WhereComposition(Where):
    """WhereComposition base class for compositions os Where objects."""
    _operator = "????"

    def __init__(self, *wheres: Union[Where, 'WhereComposition']):
        self._wheres = wheres

    def __str__(self) -> str:
        return f'({f" {self.__class__._operator} ".join(str(w) for w in self._wheres)})'


class WhereAND(WhereComposition):
    """Compose two Where objects with AND"""
    _operator = "AND"


class WhereOR(WhereComposition):
    """Compose two Where objets with OR"""
    _operator = "OR"


from typing import Any


class Queue:
    def add(self, value):
        raise NotImplementedError()

    def next(self) -> Any:
        raise NotImplementedError()

    def is_empty(self) -> bool:
        raise NotImplementedError()

    @property
    def size(self) -> int:
        raise NotImplementedError()

    @property
    def first(self) -> Any:
        raise NotImplementedError()

    @property
    def last(self) -> Any:
        raise NotImplementedError()


class Stack:
    def push(self, value):
        raise NotImplementedError()

    def pop(self) -> Any:
        raise NotImplementedError()

    def is_empty(self) -> bool:
        raise NotImplementedError()

    @property
    def top(self) -> Any:
        raise NotImplementedError()

    @property
    def size(self) -> int:
        raise NotImplementedError()


from .cpp_lib.dist.primitives import Stack, Queue


class TableNotFound(Exception):
    def __init__(self, tablename: str, basename: str):
        super().__init__(f'table {tablename} is not in {basename} database')


class TableAlreadyExists(Exception):
    def __init__(self, tablename: str, basename: str):
        super().__init__(f'table {tablename} already exists in {basename} database')

class TableIsNotReal(Exception):
    def __init__(self, tablename: str):
        super().__init__(f'table {tablename} is not real, but only Real tables are supported')


from typing import Union, Iterable
from .._internal import EqType


class PrefixTreeNode:
    def __init__(self, value: EqType = None):
        self._value = value
        self._termination = 0
        self._children = dict()
        self._child_values = set()

    def add_child(self, value: EqType) -> 'PrefixTreeNode':
        if value not in self._child_values:
            self._children[value] = (node := PrefixTreeNode(value))
            self._child_values.add(value)
        else:
            node = self.get_child(value)

        return node

    def get_child(self, value: EqType) -> 'PrefixTreeNode':
        if value in self._child_values:
            return self._children[value]
        else:
            return None

    def terminate(self):
        self._termination += 1

    @property
    def value(self) -> str:
        return self._value

    @property
    def termination(self):
        return self._termination

    def __contains__(self, item: 'PrefixTreeNode') -> bool:
        return item in self._child_values

    def __bool__(self) -> bool:
        return self.value is not None

    def __str__(self) -> str:
        return f"PrefixTreeNode[{self.value}]"

    def __eq__(self, other: 'PrefixTreeNode') -> bool:
        return self.value == other.value


class PrefixTree:
    def __init__(self):
        self._root = PrefixTreeNode()

    def add(self, sequence: Union[EqType, Iterable[EqType]]):
        if hasattr(sequence, '__iter__'):
            current_node = self._root
            for v in sequence:
                current_node = current_node.add_child(v)
            current_node.terminate()
        else:
            self._root.add_child(sequence)

    def count(self, sequence: Iterable[EqType]) -> int:
        current_node = self._root
        for v in sequence:
            if (current_node := current_node.get_child(v)) is None:
                return False
        return current_node.termination

    def contains(self, sequence: Iterable[EqType]) -> bool:
        return sequence in self

    def __contains__(self, sequence: Iterable[EqType]) -> bool:
        return self.count(sequence) > 0



from typing import Sequence, Mapping, Collection, Any
from . import _Table
from . import _Base


def base_to_dict(base: '_Base.DBase') -> dict[str, list[dict[str, Any]]]:
    tables = dict()
    for t_name in base.tables:
        t = base.table(t_name)
        tables.update({t_name: table_to_list(t)})

    return tables


def table_to_list(table: '_Table.Table') -> list[dict[str, Any]]:
    query = table[:]
    fields = [f.name for f in query.fields]
    body = query()

    rows = [dict(zip(fields, row)) for row in body]
    return rows


def table_to_schema(table: '_Table.Table') -> dict[str, str]:
    schema = dict()
    for f in table.fields.values():
        signature = f'{f.type} {"NOT NULL" if not f.is_nullable else ""} {"PRIMARY KEY" if f.is_primary else ""}'
        schema.update({f.name: signature})

    for f in table.foreign_keys.values():
        key = f'foreign key({f.slave_field.name})'
        signature = f'references {f.master_field.table.name}({f.master_field.name})'
        schema.update({key: signature})

    return schema


def map_to_base(db_name: str, data_map: Mapping[str, Sequence[Mapping[str, Any]]],
                primary_fields: Mapping[str, str] = None, foreign_fields: Mapping[str, str] = None) -> '_Base.DBase':
    """
    Create DBase from data mapping.

    data mapping entry must have following structure:
        Table_name: <Sequence of instance mappings <Field_name: Instance_value>>

    :param db_name: database filename
    :param data_map: data mapping
    :param primary_fields: mapping for spec primary key fields <Table_name: Primaryfield_name>
    :param foreign_fields: mapping for spec foreign key fields <Slave_table.Field_name: Master_table.Field_name>
    """
    db = _Base.DBase(db_name)
    primary_fields = primary_fields if primary_fields is not None else dict()
    foreign_fields = foreign_fields if foreign_fields is not None else dict()

    foreigns = dict()
    for slave, master in foreign_fields.items():
        s_table, s_field = slave.split('.')
        if s_table not in foreigns:
            foreigns[s_table] = dict()
        foreigns[s_table][s_field] = master

    for table_name, table_data in data_map.items():
        pk = primary_fields[table_name] if table_name in primary_fields else None
        table_foreigns = foreigns[table_name] if table_name in foreigns else None
        seq_to_table(db, table_name, table_data, primary_field=pk, foreign_fields=table_foreigns)

    return db


def seq_to_table(db: '_Base.DBase', name: str, rows: Sequence[Mapping[str, Any]], primary_field: str = None,
                 foreign_fields: Mapping[str, str] = None, ignore_fields: Collection[str] = None) -> '_Table.Table':
    """
    Create new Table in existing DBase from data mapping sequence.

    Each map in sequence - is a row instance.
    Row instance map entry must have following structure:
        Field_name: Instance_vale_for_field

    :param db: DBase object instance
    :param name: name for the new table
    :param rows: sequence of data mappings containing row instances for table
    :param primary_field: name of field which will be primary
    :param foreign_fields: map for foreign keys in table <Field_name: Master_table.Field_name>
    :param ignore_fields: sequence of field names to ignore
    """
    foreign_fields = foreign_fields if foreign_fields is not None else dict()
    ignore_fields = ignore_fields if ignore_fields is not None else tuple()

    schema = seq_to_schema(rows, primary_field=primary_field, ignore_fields=ignore_fields, foreign_fields=foreign_fields)
    table = db.new_table(name, schema)

    for row in rows:
        table.INSERT(row)

    return table


def seq_to_schema(rows: Sequence[Mapping[str, Any]], primary_field: str = None,
                  foreign_fields: Mapping[str, str] = None, ignore_fields: Collection[str] = None) -> dict:
    """
    Create Table SQL schema from data mapping sequence.

    Each map in sequence - is a row instance.
    Row instance map entry must have following structure:
        Field_name: Instance_vale_for_field

    :param rows: sequence of data mappings containing row instances for table
    :param primary_field: name of field which will be primary
    :param foreign_fields: map for foreign keys in table <Field_name: Master_table.Field_name>
    :param ignore_fields: sequence of field names to ignore
    """
    schema = dict()
    ignore_fields = ignore_fields if ignore_fields is not None else tuple()
    foreign_fields = foreign_fields if foreign_fields is not None else dict()

    for row in rows:
        schema.update(map_to_schema(row, primary_field=primary_field, ignore_fields=ignore_fields, foreign_fields=foreign_fields))

    return schema


def map_to_schema(row: Mapping[str, Any], primary_field: str = None,
                  foreign_fields: Mapping[str, str] = None, ignore_fields: Collection[str] = None) -> dict:
    """
    Create Table SQL schema from single data mapping.

    Given data mapping is a sample row instance.
    Row instance map entry must have following structure:
        Field_name: Instance_vale_for_field

    :param row: sequence of data mappings containing row instances for table
    :param primary_field: name of field which will be primary
    :param foreign_fields: map for foreign keys in table <Field_name: Master_table.Field_name>
    :param ignore_fields: sequence of field names to ignore
    """
    foreign_fields = foreign_fields if foreign_fields is not None else dict()
    ignore_fields = ignore_fields if ignore_fields is not None else tuple()

    schema = dict()
    if primary_field is None:
        schema.update({'id': 'INTEGER PRIMARY KEY AUTOINCREMENT'})
    else:
        schema.update({primary_field: f'{python_to_sql_type(row[primary_field])} PRIMARY KEY'})
        ignore_fields += primary_field,

    fields = list(row.keys())
    types = [python_to_sql_type(v) for v in row.values()]

    for f, t in zip(fields, types):
        if (f not in ignore_fields) and (f not in schema):
            schema.update({f: t})

    for slave, master in foreign_fields.items():
        if slave in schema.keys():
            m_table, m_field = master.split('.')
            schema.update({f'foreign key({slave})': f'references {m_table}({m_field})'})
        else:
            raise KeyError(f'foreign field field {slave} was not in table schema')

    return schema


def python_to_sql_type(value) -> str:
    """
    Convert Python type to an SQL type name.

    :param value: sample of value to extract type from it.
    """
    if isinstance(value, str):
        return "TEXT"
    elif isinstance(value, int):
        return "INTEGER"
    elif isinstance(value, float):
        return "REAL"
    elif isinstance(value, bytes):
        return "BLOB"
    elif value is None:
        return "NULL"
    raise TypeError(f"Not supported type {type(value)}")



from . import _Table


class Aggregate:
    """Base class for aggregate functions"""
    function = "???"

    def __init__(self, field_name: str):
        self.field_name = field_name
        self.field = None
        self.table = None

    @property
    def is_ready(self) -> bool:
        return (self.field is not None) and (self.table is not None)

    def compile(self, table: '_Table.Table') -> '_Table.CalculatedField':
        self.table = table
        self.field = table.field_by_name(self.field_name)
        return _Table.CalculatedField(self.field, self.function)


class COUNT(Aggregate):
    """COUNT aggregate function"""
    function = "COUNT"


class AVG(Aggregate):
    """AVG aggregate function"""
    function = "AVG"


class SUM(Aggregate):
    """SUM aggregate funtcion"""
    function = "SUM"


class MIN(Aggregate):
    """MIN aggregate function"""
    function = "MIN"


class MAX(Aggregate):
    """MAX aggregate function"""
    function = "MAX"


from . import _Graph
from typing import Hashable
from math import inf as INF


class PyOrientedGraph(_Graph.PyGraph):
    def add_edge(self, value_from: Hashable, value_to: Hashable, weight: float = 1):
        self.add_vertex(value_from)
        self.add_vertex(value_to)

        self._graph_adj[value_from].add(value_to)
        self._adj_weights[(value_from, value_to)] = weight

    def edge_weight(self, value_from: Hashable, value_to: Hashable) -> float:
        if self.are_adjacent(value_from, value_to):
            return self._adj_weights[(value_from, value_to)]
        else:
            return INF

    def are_adjacent(self, value1: Hashable, value2: Hashable) -> bool:
        if value1 in self and value2 in self:
            return value2 in self._graph_adj[value1]
        else:
            raise ValueError(f"`{value1}` or `{value2}` not in graph")


from typing import Union, Iterable

from .._internal import CompType


class BinarySubtree:
    def __init__(self, value: CompType = None, parent: 'BinarySubtree' = None):
        self._value = value
        self._depth = 0
        self._parent = parent
        self._left_child: 'BinarySubtree' = None
        self._right_child: 'BinarySubtree' = None

        if value is not None:
            self._left_child = BinarySubtree(parent=self)
            self._right_child = BinarySubtree(parent=self)
            self._depth = 1

    def goes_left(self, value: CompType) -> bool:
        if self and (value is not None):
            return self > value
        else:
            return False

    def goes_right(self, value: CompType) -> bool:
        if self and (value is not None):
            return self < value
        else:
            return False

    def set_child(self, other: 'BinarySubtree') -> 'BinarySubtree':
        was_child = None
        if self.goes_left(other):
            was_child = self._left_child
            self._left_child = other
            other.set_parent(self)
        elif self.goes_right(other):
            was_child = self._right_child
            self._right_child = other
            other.set_parent(self)

        return was_child

    def set_parent(self, other: 'BinarySubtree') -> 'BinarySubtree':
        was_parent = self._parent
        self._parent = other
        return was_parent

    def add(self, value: CompType) -> 'BinarySubtree':
        result = None
        if self:
            if self.goes_left(value):
                if self.left:
                    result = self.left.add(value)
                else:
                    result = BinarySubtree(value)
                    self.set_child(result)
            elif self.goes_right(value):
                if self.right:
                    result = self.right.add(value)
                else:
                    result = BinarySubtree(value)
                    self.set_child(result)
        else:
            self.__init__(value, self._parent)
            result = self

        if result:
            self._depth = max(self.left_subtree_depth, self.right_subtree_depth)+1

        if abs(self.left_subtree_depth - self.right_subtree_depth) > 1:
            self.balance()

        return result

    def balance(self):
        if not self:
            return
        dleft = self.left_subtree_depth
        dright = self.right_subtree_depth
        dleft_left = self.left.left_subtree_depth
        dleft_right = self.left.right_subtree_depth
        dright_left = self.right.left_subtree_depth
        dright_right = self.right.right_subtree_depth

        z = self
        p = self._parent

        if (dleft > dright) and (dleft_left >= dleft_right):
            y = self.left
            t3 = self.left.right

            p.set_child(y)
            y._right_child = z
            z._parent = y
            z._left_child = t3
            t3._parent = z

        elif (dleft > dright) and (dleft_left <= dleft_right):
            y = self.left
            x = y.right
            t2 = x.left

            z._left_child = x
            x._parent = z
            x._left_child = y
            y._parent = x
            y._right_child = t2
            t2._parent = y

        elif (dleft < dright) and (dright_left <= dright_right):
            y = self.right
            t2 = y.left

            p.set_child(y)
            y._left_child = z
            z._parent = y
            z._right_child = t2
            t2._parent = z

        elif (dleft < dright) and (dright_left >= dright_right):
            y = self.right
            x = y.left
            t3 = x.right

            z._right_child = x
            x._parent = z
            x._right_child = y
            y._parent = x
            y._left_child = t3
            t3._parent = y

            t2 = x.left
            p.set_child(x)
            x._left_child = z
            z._parent = x
            z._right_child = t2
            t2._parent = z

    def travel(self) -> Iterable['BinarySubtree']:
        if self:
            yield self
            if self.right:
                yield from self.right.travel()
            if self.left:
                yield from self.left.travel()

    @property
    def left_subtree_depth(self) -> int:
        if self.left:
            return self.left.depth
        else:
            return 0

    @property
    def right_subtree_depth(self) -> int:
        if self.right:
            return self.right.depth
        else:
            return 0

    @property
    def depth(self) -> int:
        return self._depth

    @property
    def value(self):
        return self._value

    @property
    def left(self) -> 'BinarySubtree':
        return self._left_child

    @property
    def right(self) -> 'BinarySubtree':
        return self._right_child

    @property
    def graph_string(self):
        if self:
            return f"{self}:\nl:{self.left.graph_string}\nr:{self.right.graph_string}"
        else:
            return ""

    def __contains__(self, item: CompType) -> bool:
        if self == item:
            result = True
        elif self.goes_right(item) and self.right:
            result = item in self.right
        elif self.goes_left(item) and self.left:
            result = item in self.left
        else:
            result = False

        return result

    def __iter__(self) -> Iterable['BinarySubtree']:
        return self.travel()

    def __bool__(self):
        return self.value is not None

    def __eq__(self, other: CompType | 'BinarySubtree'):
        item = other.value if isinstance(other, BinarySubtree) else other
        return self.value == item

    def __gt__(self, other: CompType | 'BinarySubtree'):
        item = other.value if isinstance(other, BinarySubtree) else other
        return self.value > item

    def __lt__(self, other: CompType | 'BinarySubtree'):
        item = other.value if isinstance(other, BinarySubtree) else other
        return self.value < item

    def __str__(self):
        return f"BinarySubtree[{self.value}]"


class BinaryTree:
    def __init__(self, value: CompType = None):
        self._child: 'BinarySubtree' = BinarySubtree(value, parent=self) if value else BinarySubtree(parent=self)
        self._count = 1 if value else 0

    def set_child(self, child: 'BinarySubtree') -> 'BinarySubtree':
        was_child = self._child
        self._child = child
        child.set_parent(self)

        return was_child

    def set_parent(self, other: 'BinarySubtree') -> 'BinarySubtree':
        raise NotImplementedError("BinaryTree can not have parent")

    def add(self, value: CompType) -> 'BinarySubtree':
        result = self._child.add(value)
        if result:
            self._count += 1
        return result

    def balance(self):
        self._child.balance()

    def travel(self) -> Iterable['BinarySubtree']:
        yield from self._child.travel()

    @property
    def depth(self):
        return self._child.depth

    @property
    def _left_child(self):
        return self._child

    @property
    def _right_child(self):
        return self._child

    @property
    def count(self):
        return self._count

    def contains(self, item: CompType) -> bool:
        return item in self

    def __contains__(self, item: CompType):
        return item in self._child

    def __iter__(self) -> Iterable['BinarySubtree']:
        return self.travel()

    def __str__(self):
        return f"BinaryTree[depth: {self.depth}, count: {self.count}]"



from typing import Iterable, Union, Mapping, Any, Sequence
from . import _Base
from . import _Query
from . import _FieldConstraints as _Constr
from . import aggregate


class Table:
    """
    Table object which represents SQL table or SQL composition of Table objects.
    """
    def __init__(self, table_name: str, db_obj: '_Base.DBase', table_query: str = None, binded_tables: Iterable['Table'] = None):
        self._name = table_name
        self._query = table_name if table_query is None else table_query
        self._is_real = binded_tables is None
        self._db = db_obj
        self._binded = {self} if self.is_real else set()
        if binded_tables is not None:
            for rqt in binded_tables:
                self._binded.update(rqt.binded)

        self._fields = dict()
        self._foreign_keys = dict()
        for rqt in self.binded:
            self._fields.update(db_obj.table_fields(rqt))
            self._foreign_keys.update(db_obj.table_foreign_keys(rqt))

        self._foreign_tables = set()
        for f, fk in self.foreign_keys.items():
            self._foreign_tables.add(fk.master_field.table)

        for field in self.fields.values():
            self.__dict__.update({f'f_{field.name}': field})

    def field_by_name(self, field_name: str) -> 'TableField':
        """
        Get TableField reference by field_name from this Table.

        :param field_name: field name with (or without) real Table name reference.
        :return:
        """
        return self._fields[self.field_from_name(field_name)]

    def field_from_name(self, field_name: str) -> str:
        """
        Copmile <Table_Name>.<Field_Name> from only Field_Name for one of this Table's real binded Tabled.

        If field is not present in Table, raise KeyError.

        :param field_name: field name.
        :return: field name with this Table's name reference
        """
        if "." in field_name:
            return field_name
        for try_table in self.binded:
            if self.has_field(fname := f'{try_table.name}.{field_name}'):
                return fname
        raise KeyError(f"{self} has no field {field_name}")

    def has_field(self, field_name: str) -> bool:
        """
        Is this Table has field.

        :param field_name: field name with (or without) real Table name reference.
        :return: This Table contains field
        """
        if "." in field_name:
            return field_name in self.fields
        else:
            for try_table in self.binded:
                if (fname := f'{try_table.name}.{field_name}') in self.fields:
                    return True
        return False

    def has_fields(self, field_names: Iterable[str]) -> bool:
        """
        Table.has_field() for Iterable of field names.

        :param field_names: Iterable of field_names.
        :return: True - all given field names are present in this Table.
        """
        for f in field_names:
            if not self.has_field(f):
                return False
        return True

    @property
    def is_real(self) -> bool:
        """
        Is this Table really contained in DBase or just a complex structure of other Tables.

        :return: True - Table is present in DBase, False - Table is a composition of real Tables.
        """
        return self._is_real

    @property
    def name(self) -> str:
        """Current Table's name. Not supposed to match SQL table name."""
        return self._name

    @property
    def query(self) -> str:
        """
        This Table query for DBase.

        This is SQL command used to get current Table's structure from DBase.

        :return: str of query reference
        """
        return self._query

    @property
    def binded(self) -> set['Table']:
        """
        All tables binded with this Table.

        'Binded' means that 'binded table' is present in DBase as real table and is used to make queries to this Table.

        :return: set of binded Tables
        """
        return self._binded

    @property
    def db(self) -> '_Base.DBase':
        """Database which contains this table."""
        return self._db

    @property
    def fields(self) -> dict['TableField']:
        """All fields of current Table."""
        return self._fields.copy()

    @property
    def foreign_keys(self) -> dict[str, 'TableFK']:
        """All foreign keys of current Table."""
        return self._foreign_keys

    @property
    def foreign_tables(self) -> set['Table']:
        """All Tables to whom current Table is connected by foreign keys."""
        return self._foreign_tables.copy()

    def catch_fk_connection(self, other: 'Table') -> 'TableFK':
        """
        Find Foreign Key that connects two tables.

        :param other: other Table.
        :return: TableFK that represents connection between this Table and other Table.
        """
        if maybe := self.binded.intersection(other.foreign_tables):
            master = self
            slave = other
        elif maybe := other.binded.intersection(self.foreign_tables):
            master = other
            slave = self
        else:
            raise KeyError(f"{self} and {other} are not connected by foreign key")

        for fk, fkey in slave.foreign_keys.items():
            try_master_ref = f'{fkey.master_field.table.name}.{fkey.master_field.name}'
            if master.has_field(try_master_ref):
                return fkey

    def join(self, other: 'Table', join: str, self_field: 'TableField', other_field: 'TableField') -> 'Table':
        """
        Make new Table, that is composition of current and other Tables, JOINED by fields.
        :param other: other Table to JOIN
        :param join: JOIN type: "INNER", "LEFT" etc
        :param self_field: field of current Table to use in JOIN comparison
        :param other_field: field of other Table to use in JOIN comparison
        :return:
        """
        name = f"{self.name}_{join}_{other.name}"

        binded = self.binded.union(other.binded)
        query = f'{self.query} {join} JOIN {other.query} ON {self_field.full_name} = {other_field.full_name}'

        return Table(name, self.db, table_query=f'({query})', binded_tables=binded)

    def DROP(self) -> None:
        """
        Drop current table from DBase.
        """
        self._db.drop(self.name)

    def INSERT(self, values: Mapping[str, Any] | Sequence):
        """
        Insert new row into this Table with specified values.

        Gaps in values are not supported.
        Only Real Tables supported.

        if <tuple> is passed:
        - Values order must match this Table fields order. See Table.fields property.
        - If Table Primary Key is AUTOINCREMENT you have not to use dict.

        if <dict> is passed:
        - keys are field names, values are values.

        :param values: <tuple | dict> definition for new row.
        """
        if not self.is_real:
            raise TypeError(f'Only Real Tables supported. {self} is a query composition of {self.binded} tables')

        if isinstance(values, tuple):
            fields = tuple(field.name for field in tuple(self._fields.values())[:len(values)])
        elif isinstance(values, dict):
            fields = tuple(values.keys())
            values = tuple(values.values())
        else:
            raise TypeError(f'expected <tuple> or <dict>, got {type(values)}')

        for f in fields:
            if not self.has_field(f):
                raise KeyError(f'Table {self} does not have field {f}')

        self.db.insert(self.name, fields, values)

    def SELECT(self, field_names: Union[slice, tuple, str, 'aggregate.Aggregate']) -> '_Query.SelectQuery':
        """
        Create SelectQuery for Table.

        :param field_names: Table field names to select
        :return: new SelectQuery
        """
        if isinstance(field_names, slice):
            field_names = tuple(self.fields.keys())
        if not isinstance(field_names, tuple):
            field_names = (field_names,)

        fields = []
        for f in field_names:
            if isinstance(f, aggregate.Aggregate):
                fields.append(f.compile(self))
            else:
                fields.append(self.field_by_name(f))

        return _Query.SelectQuery(self, fields)

    def AND(self, other: 'Table') -> 'Table':
        """
        Make cartisian prodyts of tables.

        :param other: other Table for production
        :return: new Table with is_real=False property
        """
        name = f'{self.name}_x_{other.name}'
        binded = self.binded.union(other.binded)
        query = f'({self.query} CROSS JOIN {other._query})'

        return Table(name, self._db, table_query=query, binded_tables=binded)

    def INNER(self, other: 'Table') -> 'Table':
        """
        INNER JOIN two tables by foreign key.

        :param other: other Table to join
        :return: new Table with is_real=False property
        """
        fk_connection = self.catch_fk_connection(other)

        master = fk_connection.master_field.table
        slave = fk_connection.slave_field.table

        master_ref = fk_connection.master_field
        slave_ref = fk_connection.slave_field

        return self.join(other, 'INNER', master_ref, slave_ref)

    def LEFT(self, other: 'Table') -> 'Table':
        """
        LEFT JOIN two tables by foreign key.

        If you want a RIGHT JOIN, just LEFT JOIN tables in reversed order.

        :param other: other Table to exclude from this Table.
        :return:new Table with is_real=False property
        """
        fk_connection = self.catch_fk_connection(other)

        if fk_connection.master_field.table == self:
            self_ref = fk_connection.master_field
            other_ref = fk_connection.slave_field
        else:
            self_ref = fk_connection.slave_field
            other_ref = fk_connection.master_field

        return self.join(other, 'LEFT', self_ref, other_ref)

    def FULL(self, other: 'Table') -> 'Table':
        """
        FULL JOIN two tables by foreign key.

        :param other: other Tabl to join.
        :return: new Table with is_real=False property
        """
        fk_connection = self.catch_fk_connection(other)
        master = fk_connection.master_field.table
        slave = fk_connection.slave_field.table

        master_ref = fk_connection.master_field
        slave_ref = fk_connection.slave_field

        return self.join(other, 'FULL', master_ref, slave_ref)

    def __getitem__(self, field_names: Union[slice, tuple, str, 'aggregate.Aggregate']) -> '_Query.SelectQuery':
        """
        Create SelectQuery for Table.

        :param field_names: Table field names to select
        :return: new SelectQuery
        """
        return self.SELECT(field_names)

    def __mul__(self, other: 'Table') -> 'Table':
        """
        Make cartisian prodyts of tables.

        :param other: other Table for production
        :return: new Table with is_real=False property
        """
        return self.AND(other)

    def __xor__(self, other: 'Table') -> 'Table':
        """
        FULL JOIN two tables by foreign key.

        :param other: other Tabl to join.
        :return: new Table with is_real=False property
        """
        return self.FULL(other)

    def __sub__(self, other: 'Table') -> 'Table':
        """
        LEFT JOIN two tables by foreign key.

        If you want a RIGHT JOIN, just LEFT JOIN tables in reversed order.

        :param other: other Table to exclude from this Table.
        :return:new Table with is_real=False property
        """
        return self.LEFT(other)

    def __and__(self, other: 'Table') -> 'Table':
        """
        INNER JOIN two tables by foreign key.

        :param other: other Table to join
        :return: new Table with is_real=False property
        """
        return self.INNER(other)

    def __lshift__(self, values: Mapping[str, Any] | Sequence) -> None:
        """
        Insert new row into this Table with specified values.

        Gaps in values are not supported.
        Only Real Tables supported.

        if <tuple> is passed:
        - Values order must match this Table fields order. See Table.fields property.
        - If Table Primary Key is AUTOINCREMENT you have not to use dict.

        if <dict> is passed:
        - keys are field names, values are values.

        :param values: <tuple | dict> definition for new row.
        """
        self.INSERT(values)

    def __repr__(self) -> str:
        return f'Table<{self.name} of {self.db.name}, {"Real" if self.is_real else "Composition"}>'

    def __eq__(self, other: 'Table'):
        return (self.query == other.query) and (self.db.name == other.db.name)

    def __hash__(self):
        return hash((self.query, self.db.name))


class TableField:
    """TableField object, that represents some Table field."""
    def __init__(self, i: int, name: str, typ: str, table_obj: Table,
                 constraints: Iterable['_Constr.FieldConstraint'] = None):
        self.id = i
        self.name = name
        self.type = typ
        self.table = table_obj
        self.constraints = list(constraints) if constraints else list()

    @property
    def full_name(self) -> str:
        """Full name means <Table name>.<Field name>"""
        return self.table.field_from_name(self.name)

    @property
    def is_primary(self) -> bool:
        for c in self.constraints:
            if isinstance(c, _Constr.Primary):
                return True
        return False

    @property
    def is_nullable(self) -> bool:
        for c in self.constraints:
            if isinstance(c, _Constr.NotNull):
                return False
        return True

    @property
    def is_foreign(self) -> bool:
        for c in self.constraints:
            if isinstance(c, _Constr.Foreign):
                return True
        return False

    @property
    def is_checked(self) -> bool:
        for c in self.constraints:
            if isinstance(c, _Constr.Check):
                return True
        return False

    def __repr__(self) -> str:
        return f'Field<{self.id}, {self.name}, {self.type} of {self.table}, {self.constraints}>'

    def __eq__(self, other: 'TableField') -> bool:
        return (self.name == other.name) and (self.table == other.table)

    def __hash__(self):
        return hash((self.name, self.table.name, self.table.db.name))


class CalculatedField:
    def __init__(self, field: 'TableField', function: str):
        self.field = field
        self.function = function

    @property
    def full_name(self) -> str:
        return f'{self.function}({self.field.full_name})'

    @property
    def id(self) -> int:
        return self.field.id

    @property
    def table(self) -> 'Table':
        return self.field.table

    @property
    def type(self) -> str:
        return self.field.type

    @property
    def name(self):
        return self.field.name

    def __repr__(self) -> str:
        return f'CalculatedField<{self.function} from {self.field}>'

    def __eq__(self, other: Union['CalculatedField', 'TableField']) -> bool:
        equality = (self.name == other.name) and (self.table == other.table)
        if isinstance(other, CalculatedField):
            equality = equality and (self.function == other.function)

        return equality

    def __hash__(self):
        return hash((self.name, self.table.name, self.table.db.name, self.function))


class TableFK:
    """TableFK object represents connection between two TableFields by foreign key constraint."""
    def __init__(self, master_field: TableField, slave_field: TableField):
        self._master_field = master_field
        self._slave_field = slave_field

    @property
    def master_field(self) -> 'TableField':
        """Field of table which is Master in foreign key connection."""
        return self._master_field

    @property
    def slave_field(self) -> 'TableField':
        """Field of table which is Slave in foreign key connection."""
        return self._slave_field

    def __repr__(self) -> str:
        return f'TableFK<{self.slave_field} -> {self.master_field}>'

    def __eq__(self, other: 'TableFK'):
        return (self.master_field == other.master_field) and (self.slave_field == other.slave_field)

    def __hash__(self):
        return hash((self.master_field, self.slave_field))